// Mike Woodring
// http://staff.develop.com/woodring
//
// This sample documents & demonstrates a couple of programmatic interfaces
// that let you enumerate the contents of the global, download, or native
// assembly caches.
//
// Command line usage:
//
//   enumcache (/gac | /dl | /native)+ (/ver /token /culture /key)*
//
// The interface definitions (or in some cases, stub definitions)
// defined below in the AssemblyCache namespace were created based on
// information available in the Rotor sources (specifically, fusion.idl).
//
// A related sample ships as part of the .NET Framework SDK that demonstrates
// programmatically installing and removing assemblies from the global
// assembly cache.  See <.net sdk>\samples\technologies\interop\advanced\comreg
// (specifically, fusioninstall.cs) for details.
//
// Known issues:
// * Specifying ASM_DISPLAYF_PUBLIC_KEY in the call to GetDisplayName always results in PublicKey=null.
// * Have not figured out how to enumerate more than one cache (global, download, native) using
//   just one enumerator.  Bitwise ORing the CacheFlags together in the obvious manner results
//   in only the interface to the download cache being returned (that I can tell at least).
//
using System;
using System.Text;
using System.Runtime.InteropServices;
using System.Reflection;

namespace AssemblyCache
{
    [
      ComImport,
      Guid("21b8916c-f28e-11d2-a473-00c04f8ef448"),
      InterfaceType(ComInterfaceType.InterfaceIsIUnknown)
    ]
    public interface IAssemblyEnum
    {
        [PreserveSig]
        int GetNextAssembly( IntPtr reservedPassZero, out IAssemblyName asmName, uint flagsPassZero );

        void Reset();
        IAssemblyEnum Clone();
    };

    [
      ComImport,
      Guid("CD193BC0-B4BC-11d2-9833-00C04FC31D2E"),
      InterfaceType(ComInterfaceType.InterfaceIsIUnknown)
    ]
    public interface IAssemblyName
    {
        void SetProperty_StubOnly_InvalidArgList_Do_Not_Call();
        void GetProperty_StubOnly_InvalidArgList_Do_Not_Call();
        void Finalize();
        void GetDisplayName( StringBuilder displayName, ref uint cch, NameDisplayFlags flags );
        void BindToObject_StubOnly_InvalidArgList_Do_Not_Call();
        void GetName( ref uint cch, StringBuilder name );
        void GetVersion( out uint versionHi, out uint versionLow );
        void IsEqual( IAssemblyName pName, uint cmpFlags );
        IAssemblyName Clone();
    }

    [Flags]
    public enum NameDisplayFlags : uint
    {
        ASM_DISPLAYF_VERSION    = 0x1,
        ASM_DISPLAYF_CULTURE    = 0x2,
        ASM_DISPLAYF_PUBLIC_KEY_TOKEN = 0x4,
        ASM_DISPLAYF_PUBLIC_KEY = 0x8,
        ASM_DISPLAYF_CUSTOM     = 0x10,
        ASM_DISPLAYF_PROCESSORARCHITECTURE = 0x20,
        ASM_DISPLAYF_LANGUAGEID     = 0x40
    }

    [Flags]
    public enum CacheFlags : uint
    {
        ASM_CACHE_ZAP = 0x1,            // NGEN'd assemblies.
        ASM_CACHE_GAC = 0x2,            // Global assembly cache.
        ASM_CACHE_DOWNLOAD = 0x4        // Download cache.
    };

    public class Controllers
    {
        [DllImport("Fusion.dll", CharSet=CharSet.Auto, EntryPoint="CreateAssemblyEnum")]
        public static extern int GetEnumerator( out IAssemblyEnum ppAsmCache,
                                                IntPtr pAppCtx, IntPtr pName,
                                                CacheFlags dwFlags, IntPtr pvReserved );
    }
}

class CacheEnum
{
    static void Main( string[] args )
    {
        // Usage: enumcache (/gac | /dl | /native)+ (/ver /token /culture /key)*
        //
        int hr;
        AssemblyCache.CacheFlags cacheFlags = 0;
        AssemblyCache.NameDisplayFlags nameFlags = 0;

        // Check for command line flags.
        //
        foreach( string arg in args )
        {
            switch( arg )
            {
                // Only one of these can be specified at this time
                // (last one specified on the command line wins)...
                //
                case "/gac": cacheFlags = AssemblyCache.CacheFlags.ASM_CACHE_GAC; break;
                case "/dl": cacheFlags = AssemblyCache.CacheFlags.ASM_CACHE_DOWNLOAD; break;
                case "/native": cacheFlags = AssemblyCache.CacheFlags.ASM_CACHE_ZAP; break;

                // Along with one or more of these.
                //
                case "/ver": nameFlags |= AssemblyCache.NameDisplayFlags.ASM_DISPLAYF_VERSION; break;
                case "/key": nameFlags |= AssemblyCache.NameDisplayFlags.ASM_DISPLAYF_PUBLIC_KEY; break;
                case "/token": nameFlags |= AssemblyCache.NameDisplayFlags.ASM_DISPLAYF_PUBLIC_KEY_TOKEN; break;
                case "/culture": nameFlags |= AssemblyCache.NameDisplayFlags.ASM_DISPLAYF_CULTURE; break;
            }
        }

        if( cacheFlags == 0 )
        {
            // Default to the GAC if nothing is specified on the command line.
            //
            cacheFlags = AssemblyCache.CacheFlags.ASM_CACHE_GAC;
        }

        AssemblyCache.IAssemblyEnum gacEnum;
        
        hr = AssemblyCache.Controllers.GetEnumerator( out gacEnum, IntPtr.Zero, IntPtr.Zero,
                                                      cacheFlags, IntPtr.Zero );

        if( hr != 0 ) // S_OK
        {
            Console.WriteLine("Failed to acquire enumerator interface.  HR = 0x{0:x}.", hr);

            if( hr == 1 ) // S_FALSE
            {
                Console.WriteLine("It's likely that one of the specified caches has zero items in it at this time.");
            }

            return;
        }

        AssemblyCache.IAssemblyName asmNameObj = null;
        int nAssemblies = 0;
        StringBuilder asmName = new StringBuilder(256);

        do
        {
            // Get the assembly name object representing the next
            // item in the specified part of the assembly cache.
            //
            hr = gacEnum.GetNextAssembly(IntPtr.Zero, out asmNameObj, 0);

            if( hr == 0 )
            {
                nAssemblies++;

                if( nameFlags == 0 )
                {
                    // This code path mostly to demonstrate how to parse the
                    // assembly version.
                    //
                    uint cch = (uint)asmName.Capacity;
                    asmNameObj.GetName(ref cch, asmName);

                    uint verHi;
                    uint verLow;
                    asmNameObj.GetVersion(out verHi, out verLow);

                    ushort major = (ushort)(verHi >> 16);
                    ushort minor = (ushort)(verHi & 0xffff);
                    ushort build = (ushort)(verLow >> 16);
                    ushort revision = (ushort)(verLow & 0xffff);

                    // Display the current assembly's name and version information.
                    //
                    Console.WriteLine( "{0}, Version={1}.{2}.{3}.{4}",
                                    asmName, major, minor, build, revision );
                }
                else
                {
                    // Have the assembly name object do its own formatting.
                    //
                    uint cch = (uint)asmName.Capacity;
                    asmNameObj.GetDisplayName(asmName, ref cch, nameFlags);
                    Console.WriteLine(asmName);
                }

                // Release the assembly name object now rather
                // than waiting for a GC cycle.
                //
                Marshal.ReleaseComObject(asmNameObj);
                asmNameObj = null;
            }
        }
        while( hr == 0 );

        Console.WriteLine();
        Console.WriteLine("{0} {1} displayed.", nAssemblies, (nAssemblies == 1 ? "assembly" : "assemblies"));
    }
}
